import logging

import numpy as np
from numpy.linalg import norm

from lsqAxSI.matrix_operations import normalize, compute_w

logger = logging.getLogger(__name__)


class Precondition():
    def __init__(self, H, DM, DG, grad):
        self.H = H
        self.DM = DM
        self.DG = DG
        self.r = -grad
        self.rshape = self.r.shape
        self.v1 = np.zeros(self.rshape)
        self.it = 0
        self.posdef = 1
        self.inner1 = 0
        self.inner2 = 0

    def init_procedure(self, ppvec, RPCMTX, tol):
        self.compute_vars(0, ppvec, RPCMTX)
        self.stoptol = tol * norm(self.gradient_z)

    def solve(self, kmax, ppvec, RPCMTX, tol):
        self.init_procedure(ppvec, RPCMTX, tol)

        # Primary loop
        for k in range(max(kmax, 1)):
            if self.conjugate_gradients_break():
                break
            self.conjugate_gradients_advance(k, ppvec, RPCMTX)
            # Exit?
            if norm(self.gradient_z) <= self.stoptol:
                break

        self.v1 = normalize(self.v1)
        self.it = k

    def conjugate_gradients_break(self):
        if self.denom <= 0:
            self.posdef = 0
            self.v1 = self.d
            return True
        return False

    def conjugate_gradients_advance(self, k, ppvec, RPCMTX):
        alpha = self.inner1 / self.denom
        self.v1 = self.v1 + alpha * self.d
        self.r = self.r - alpha * self.ww
        self.compute_vars(k, ppvec, RPCMTX)

    def compute_vars(self, k, ppvec, RPCMTX):
        self.gradient_z = self.apply_preconditioner_to_vector(ppvec, RPCMTX)
        self.compute_d(k)
        w = compute_w(self.H, self.DM, self.d)
        self.ww = w + np.dot(self.DG, self.d)
        self.denom = np.dot(self.d.T, self.ww)
        self.inner2 = self.inner1
        self.inner1 = np.dot(self.r.T, self.gradient_z)

    def compute_d(self, k):
        if k == 0:
            self.d = self.gradient_z
        else:
            beta = self.inner1 / self.inner2
            self.d = self.gradient_z + beta * self.d

    def apply_preconditioner_to_vector(self, ppvec, RPCMTX):
        # initialization
        w = np.zeros(self.rshape)
        if ppvec.size == 0:
            ppvec = np.arange(self.rshape[0])
            if RPCMTX.size == 0:
                RPCMTX = np.eye(self.rshape[0])

        # Precondition
        wbar = np.linalg.solve(RPCMTX.T, self.r[ppvec])
        w[ppvec] = np.linalg.solve(RPCMTX, wbar)

        return w



import numpy as np

import scipy.sparse as sps
import logging

logger = logging.getLogger(__name__)

EPS = np.finfo('float').eps
EMPTY_ARRAY = np.array([])

class Aprecon():
    def __init__(self, H, DM, DG, n):
        self.H = H
        self.DM = DM
        self.DG = DG
        self.n = n
        self.ppvec = EMPTY_ARRAY
        self.RPCMTX = EMPTY_ARRAY
        # Form matrix m
        self.TM = np.dot(self.H, self.DM)

    def determine_factor(self, pcflags):
        ''' In our case pcflags is preset to inf'''
        # Determin factor of preconditioner
        if not pcflags:  # Diag
            self.diag_preconditioner_factor()
        elif pcflags >= self.n - 1:
            self.singularity_preconditioner_factor()
        else:
            print('aprecon: InvalidUpperBandw')

    def diag_preconditioner_factor(self):
        TM = self.TM
        M = np.dot(TM.T, TM) + self.DG
        dnrms = np.sqrt(np.sum(M * M)).T
        d = max(np.sqrt(dnrms), np.sqrt(EPS))
        self.RPCMTX = d
        self.ppvec = np.arange(self.n)

    def singularity_preconditioner_factor(self):
        TM = self.TM
        dgg = np.sqrt(self.DG)
        TM = np.vstack([TM, dgg])
        self.ppvec = np.arange(2)
        _, self.RPCMTX = np.linalg.qr(TM[:, self.ppvec])
        self.modify_for_singularity(dgg)

    def modify_for_singularity(self, dgg):
        mdiag = min(abs(np.diag(self.RPCMTX)))
        lambd = 1
        while mdiag < np.sqrt(EPS):
            TM = np.concatenate(np.dot(self.H, self.DM), dgg + lambd * np.eye(self.n), axis=1)
            self.ppvec = sps.linalg.splu(TM, perm_spec='COLAMD')
            self.RPCMTX = np.linalg.qr(TM[:, self.ppvec])
            lambd = 4 * lambd
            mdiag = min(abs(np.diag(self.RPCMTX)))



#!/usr/bin/env python

import argparse
import numpy as np
from lsqAxSI.lsq_axsi import nonlinear_least_squares
from lsqAxSI.default_values import reg_func, jac_calc


def main():
    parser = argparse.ArgumentParser(description="Perform nonlinear least-squares optimization for AxSI.")

    parser.add_argument("--x0", type=float, nargs='+', required=True,
                        help="Initial guess for the parameters (space-separated list of floats).")
    parser.add_argument("--bounds", type=float, nargs='+', required=True,
                        help="Bounds for the parameters (space-separated list of floats, format: lower1 upper1 lower2 upper2 ...).")
    parser.add_argument("--ftol", type=float, default=1e-6,
                        help="Tolerance for the cost function value (default: 1e-6).")
    parser.add_argument("--xtol", type=float, default=1e-6,
                        help="Tolerance for parameter updates (default: 1e-6).")
    parser.add_argument("--diff_step", type=float, default=1e-3,
                        help="Step size for finite-difference approximation of the Jacobian (default: 1e-3).")
    parser.add_argument("--max_nfev", type=int, default=20000,
                        help="Maximum number of function evaluations (default: 20000).")

    args = parser.parse_args()

    # Parse bounds into a tuple of arrays
    bounds_array = np.array(args.bounds)
    if len(bounds_array) % 2 != 0:
        raise ValueError("Bounds must be provided in pairs (lower and upper for each parameter).")
    lower_bounds = bounds_array[0::2]
    upper_bounds = bounds_array[1::2]
    bounds = (lower_bounds, upper_bounds)

    # Initial parameters
    x0 = np.array(args.x0)

    result = nonlinear_least_squares(reg_func=reg_func, x0=x0, bounds=bounds, jac=jac_calc, ftol=args.ftol,
                                     xtol=args.xtol, diff_step=args.diff_step, max_nfev=args.max_nfev)

    print("Optimization Result:")
    print(f"Success: {result.exitflag == 1}")
    print(f"Optimized Parameters: {result.x}")
    print(f"Residuals: {result.fvec}")
    print(f"Jacobian: {result.jacob}")
    print(f"Exit Flag: {result.exitflag}")
    print(f"Output Details: {result.output}")
    print(f"Message: {result.msg}")


if __name__ == "__main__":
    main()





import logging

import numpy as np
from numpy.linalg import norm

from lsqAxSI.lsq_aprecon import Aprecon
from lsqAxSI.lsq_components import LSQValues, LSQVars
from lsqAxSI.lsq_conjugate_gradients import Precondition
from lsqAxSI.matrix_operations import compute_w, vector_len, normalize, atamult

logger = logging.getLogger(__name__)

EPS = np.finfo('float').eps
EMPTY_ARRAY = np.array([])
INF = np.inf


class Trust_region():
    def __init__(self, values: LSQValues, vars: LSQVars):
        self.g = values.grad
        self.H = values.A
        self.ub = vars.ub
        self.lb = vars.lb

        self.DM = np.diag(np.sqrt(abs(vars.v)))
        self.DG = np.diag(abs(self.g) * vars.dv)
        self.grad = np.dot(self.DM, self.g)
        self.n = self.g.shape[0]
        self.Z = EMPTY_ARRAY
        self.v1 = vars.dnewt
        self.pcgit = 0
        self.posdef = 1
        self.exact_trust = Trust_Exact_Solution(vars.delta)

    class QP():
        def __init__(self, st, ss, s):
            self.val = INF
            self.alpha = 1
            self.st = st
            self.ss = ss
            self.s = s

        def __le__(self, other):
            ''' Compares to instances of QP '''
            return self.val <= other.val

        def multiply_vars(self, coeff=None):
            if coeff == None:
                coeff = self.alpha
            self.s = coeff * self.s
            self.ss = coeff * self.ss
            self.st = coeff * self.st

        def update_val(self, rhs, MM, addition=0):
            st = self.st
            val = np.dot(rhs.T, st) + np.dot((0.5 * st).T, np.dot(MM, st))
            self.val = val + addition

    def trial_step(self, x, theta, kmax, pcflags, pcgtol, delta):
        self.determine_subspace(x, theta, kmax, pcflags, pcgtol)
        if len(self.g) > 1:
            self.reflected_direction_evaluation(x, delta, theta)
            self.gradient_direction_evaluation(x, theta)
        self.choose_best_direction()

    def determine_subspace(self, x, theta, kmax, pcflags, tol):
        if self.v1.size == 0:
            self.call_conjugate_gradients(kmax, pcflags, tol)
        self.Z = self.v1
        if self.n > 1:
            self.extend_Z()
        self.first_qp = self.compute_qp(x, self.Z, theta, save_s=True)

    def call_conjugate_gradients(self, kmax, pcflags, tol):
        precon = Aprecon(self.H, self.DM, self.DG, self.n)
        precon.determine_factor(pcflags)

        pcgr = Precondition(self.H, self.DM, self.DG, self.grad)
        pcgr.solve(kmax, precon.ppvec, precon.RPCMTX, tol)

        self.v1 = pcgr.v1
        self.pcgit = pcgr.it
        self.posdef = pcgr.posdef

    def reduce_to_subspace(self, Z, grad):
        W = compute_w(self.H, self.DM, Z)
        self.MM = np.dot(Z.T, W) + np.dot(Z.T, np.dot(self.DG, Z))
        self.rhs = np.dot(Z.T, grad)

    def gradient_direction_evaluation(self, x, theta):
        ''' Evaluate along gradient direction '''
        ZZ = normalize(self.grad)
        self.second_qp = self.compute_qp(x, ZZ, theta)

    def compute_qp(self, x, Z, theta, save_s=False):
        self.reduce_to_subspace(Z, self.grad)
        st = self.exact_trust.compute_solution(self.rhs, self.MM)
        ss = np.dot(Z, st)
        s = abs(np.diag(self.DM)) * ss
        qp = self.QP(st, ss, s)
        self.finish_qp_evaluation(x, qp, theta)
        if save_s:
            self.set_s_values(st, ss, s)
        return qp

    def reflected_direction_evaluation(self, x, delta, theta):
        self.third_qp = self.QP(self.st, self.ss, self.s)
        self.third_qp.multiply_vars(self.mdis)
        self.keep_third_ss = self.third_qp.ss
        self.keep_thirs_s = self.third_qp.s

        if norm(self.third_qp.ss) < 0.9 * delta:
            nx = x + self.third_qp.s
            self.s[self.ipt] = -self.s[self.ipt]
            self.ss[self.ipt] = -self.ss[self.ipt]
            self.third_qp.update_val(self.rhs, self.MM)
            self.reflected_reduction()
            self.reflected_tau_evaluation(delta)
            self.finish_qp_evaluation(nx, self.third_qp, theta, addition=self.third_qp.val)

    def reflected_reduction(self):
        third_s = self.third_qp.s
        third_ss = self.third_qp.ss
        ng = atamult(self.H, third_s, 0)
        ng = ng + self.g
        ngrad = np.dot(self.DM, ng) + np.dot(self.DG, third_ss)

        ZZ = normalize(self.ss)
        self.reduce_to_subspace(ZZ, ngrad)

    def reflected_tau_evaluation(self, delta):
        tau = self.quadratic_zero_finder(self.third_qp.ss, delta)
        self.third_qp.ss = tau * self.ss
        self.third_qp.st = tau / norm(self.third_qp.ss)
        self.third_qp.s = abs(np.diag(self.DM)) * self.third_qp.ss

    def finish_qp_evaluation(self, x, qp, theta, addition=0):
        self.truncate_tr_solution(x, qp, theta)
        qp.multiply_vars()
        qp.update_val(self.rhs, self.MM, addition)

    def truncate_tr_solution(self, x, qp, theta):
        arg = abs(qp.s) > 0
        if np.any(np.isnan(qp.s)):
            print("Trust region step contains NaN's")
        # no truncation if s is zero length
        if not np.any(arg):
            self.mdis = 1
            qp.alpha = 1
            return
        self.nonzero_truncation(x, qp, arg, theta)

    def nonzero_truncation(self, x, qp, arg, theta):
        l = self.lb
        u = self.ub
        dis = np.maximum((u[arg] - x[arg]) / qp.s[arg], (l[arg] - x[arg]) / qp.s[arg])
        self.mdis = min(dis)
        self.ipt = np.argmin(dis)
        mdis = self.mdis * theta
        qp.alpha = min(1, mdis)

    def extend_Z(self):
        v1 = self.v1
        if self.posdef < 1:
            v2 = np.dot(self.DM, np.sign(self.grad))
        else:
            v2 = self.grad
        v2 = normalize(v2)
        v2 = v2 - np.dot(v1, np.dot(v1.T, v2))
        if norm(v2) > np.sqrt(EPS):
            v2 = normalize(v2)
            self.Z = np.vstack([self.Z, v2]).T

    def quadratic_zero_finder(self, ss, delta):
        x = self.ss

        a = np.dot(x.T, x)
        b = 2 * np.dot(ss.T, x)
        c = np.dot(ss.T, ss) - delta ** 2

        numer = -(b + np.sign(b) * np.sqrt(b ** 2 - 4 * a * c))
        r1 = numer / (2 * a)
        r2 = c / (a * r1)

        tau = max(r1, r2)
        tau = min(1, tau)
        if tau <= 0:
            print("square root error in trdog/quad1d")
        return tau

    def choose_best_direction(self):
        qp1 = self.first_qp
        qp2 = self.second_qp
        qp3 = self.third_qp

        if qp2 <= qp1 and qp2 <= qp3:
            self.set_values(qp2)
        elif qp1 <= qp2 and qp1 <= qp3:
            self.set_values(qp1)
        else:
            self.qpval = qp3.val
            self.s = qp3.s + self.keep_thirs_s
            self.ss = qp3.ss + self.keep_third_ss

    def set_values(self, qp):
        self.qpval = qp.val
        self.set_s_values(qp.st, qp.ss, qp.s)

    def set_s_values(self, st, ss, s):
        self.st = st
        self.ss = ss
        self.s = s


class Trust_Exact_Solution():
    def __init__(self, delta):
        self.delta = delta
        self.tol = 10 ** -12
        self.tol2 = 10 ** -8
        self.key = 0
        self.lambd = 0
        self.itbnd = 50
        self.count = 0

    def set_vars(self, g, H):
        self.evauluate_eigen(H)
        self.n = vector_len(g)
        self.coeff = np.zeros(self.n)
        self.laminit = -self.mineig
        self.alpha = -np.dot(self.eigvec.T, g)
        self.sig = np.sign(self.alpha[self.jmin]) + (self.alpha[self.jmin] == 0)

    def evauluate_eigen(self, H):
        try:
            self.eigval, self.eigvec = np.linalg.eig(H)
            self.mineig = min(self.eigval)
            self.jmin = np.argmin(self.eigval)
        except:
            self.eigval = np.full(1, H)
            self.eigvec = np.ones(1)
            self.mineig = self.eigval
            self.jmin = 0

    def compute_solution(self, g, H):
        self.set_vars(g, H)
        if self.mineig > 0:
            if self.positive_definite_case():
                return self.s
            self.laminit = 0
        self.indefinite_case()
        return self.s

    def positive_definite_case(self):
        self.coeff = self.alpha / self.eigval
        self.s = np.dot(self.eigvec, self.coeff)
        if norm(self.s) < 1.2 * self.delta:
            return True
        return False

    def indefinite_case(self):
        if self.secular_equation(self.laminit) > 0:
            # b = scipy.optimize.root_scalar(self.secular_equation, x0=self.laminit, bracket=(self.laminit, self.laminit+self.delta), max_iter=self.itbnd)
            b = self.right_find_zero()
            vval = abs(self.secular_equation(b))
            if vval <= self.tol2:
                self.state2(b)
                return
        self.state3()

    def state2(self, b, loop_flag=False):
        self.lambd = b
        s = self.compute_s(self.lambd)
        if 0.8 * self.delta < norm(s) < 1.2 * self.delta or loop_flag:
            self.s = s
            return s
        self.state3()

    def state3(self):
        self.lambd = -self.mineig
        arg = abs(self.eigval + self.lambd) < 10 * EPS * np.maximum(abs(self.eigval), np.ones(self.n))
        self.alpha[arg] = 0
        s = self.compute_s(self.lambd)
        if norm(s) < 0.8 * self.delta:
            beta = np.sqrt(self.delta ** 2 - norm(s) ** 2)
            s = s + beta * self.sig * self.eigvec.T[self.jmin]
        if norm(s) > 1.2 * self.delta:
            b = self.right_find_zero()
            s = self.state2(b, loop_flag=True)
        self.s = s

    def compute_s(self, lambd):
        w = self.eigval + lambd * np.ones(self.n)
        arg1 = (w == 0) & (self.alpha == 0)
        arg2 = (w == 0) & (self.alpha != 0)
        self.coeff[w != 0] = self.alpha[w != 0] / w[w != 0]
        self.coeff[arg1] = 0
        self.coeff[arg2] = INF
        self.coeff[np.isnan(self.coeff)] = 0
        s = np.dot(self.eigvec, self.coeff)
        return s

    def secular_equation(self, lambd):
        m = vector_len(lambd)
        n = vector_len(self.eigval)
        unn = np.ones(n)
        unm = np.ones(m)

        M = self.eigval * unm + unn * lambd
        MC = M
        MM = self.alpha * unm
        M[MC != 0] = MM[MC != 0] / M[MC != 0]
        M[MC == 0] = INF * np.ones(MC[MC == 0].shape)
        M = M * M

        value = np.sqrt(unm / np.dot(M.T, unn))
        # if value == np.NaN: # Refael:
        if value == np.nan:  # Refael
            value = 0
        value = (1 / self.delta) * unm - value

        return value

    def right_find_zero(self):
        # Initialization 
        self.init_right_find_zero()

        self.find_b()
        return self.b

    def init_right_find_zero(self):
        x = self.laminit
        self.itfun = 0
        if x != 0:
            dx = abs(x) / 2
        else:
            dx = 1 / 2

        self.init_var_right_find_zero("a", x)
        self.init_var_right_find_zero("b", x + 1)
        self.find_change_of_sign(x, dx)
        self.c = self.a
        self.fc = self.fb

    def init_var_right_find_zero(self, varname, value):
        fvalue = self.secular_equation(value)
        setattr(self, varname, value)
        setattr(self, "f" + varname, fvalue)
        self.itfun = self.itfun + 1

    def find_change_of_sign(self, x, dx):
        while (self.fa > 0) == (self.fb > 0):
            dx = 2 * dx
            self.init_var_right_find_zero("b", x + dx)
            if self.itfun > self.itbnd:
                return
        return

    def find_b(self):
        while self.fb != 0:
            if (self.fb > 0) == (self.fc > 0):
                self.c = self.a
                self.fc = self.fa
                self.d = self.b - self.a
                self.e = self.d
            if abs(self.fc) < abs(self.fb):
                self.a = self.b
                self.b = self.c
                self.c = self.a
                self.fa = self.fb
                self.fb = self.fc
                self.fc = self.fa
            if self.convergence_test():
                return self.b
            self.choose_bisection_or_interpolation()
            self.find_next_point()

    def find_next_point(self):
        self.a = self.b
        self.fa = self.b
        if abs(self.d) > self.rf_toler:
            self.b = self.b + self.d
        else:
            if self.b > self.c:
                self.b = self.b - self.rf_toler
            else:
                self.b = self.b + self.rf_toler
        self.fb = self.secular_equation(self.b)
        self.itfun = self.itfun + 1

    def convergence_test(self):
        if self.itfun > self.itbnd:
            return True
        self.m = 0.5 * (self.c - self.b)
        self.rf_toler = 2.0 * self.tol * max(abs(self.b), 1)
        if (abs(self.m) <= self.rf_toler) or (self.fb == 0.0):
            return True
        return False

    def choose_bisection_or_interpolation(self):
        if (abs(self.e) < self.rf_toler) or (abs(self.fa) <= abs(self.fb)):
            self.choose_bisection()
        self.choose_interpolation()

    def choose_bisection(self):
        self.d = self.m
        self.e = self.m

    def choose_interpolation(self):
        s = self.fb / self.fa
        if self.a == self.c:
            self.linear_interpolation(s)
        else:
            self.inverse_quadratic_interpolation(s)

    def linear_interpolation(self, s):
        p = 2.0 * self.m * s
        q = 1.0 - s
        self.test_interpolation(p, q)

    def inverse_quadratic_interpolation(self, s):
        q = self.fa / self.fc
        r = self.fb / self.fc
        p = s * (2.0 * self.m * q * (q - r) - (self.b - self.a) * (r - 1.0))
        q = (q - 1.0) * (r - 1.0) * (s - 1.0)
        self.test_interpolation(p, q)

    def test_interpolation(self, p, q):
        if p > 0:
            q = -q
        else:
            p = -p
        self.is_interpolated_point_acceptable(p, q)

    def is_interpolated_point_acceptable(self, p, q):
        if (2.0 * p < 3.0 * self.m * q - abs(self.rf_toler * q)) and p < abs(0.5 * self.e * q):
            self.e = self.d
            self.d = p / q
        else:
            self.choose_bisection()



# This file provide methods for matrix calculations

import logging

import numpy as np

logger = logging.getLogger(__name__)


def vector_len(v) -> int:
    """
    v could be either number or nd-array
    if v is instance of ndarray, return its first dimension
    else v has no shape attribute, treat it as an array of length 1
    """
    try:
        # if v is instance of nd-array
        length = v.shape[0]
    except:
        # v is a number
        length = 1
    return length


def normalize(v: np.ndarray) -> np.ndarray:
    """
    v is instance of nd-array
    returns v divided by its euclidean norm
    """
    if np.linalg.norm(v) > 0:
        return v / np.linalg.norm(v)
    return v


def atamult(A: np.ndarray, Y: np.ndarray, flag: int) -> np.ndarray:
    """ Multiply matrices A,Y according to a given flag """
    if flag == 0:
        V = np.dot(A.T, np.dot(A, Y))  # V = Trans(A) @ A @ Y
    elif flag > 0:
        V = np.dot(A, Y)  # V = A @ Y
    else:
        V = np.dot(A.T, Y)  # V = Trans(A) @ Y
    return V


def compute_w(H: np.ndarray, DM: np.ndarray, Z: np.ndarray) -> np.ndarray:
    """
    A series of matrix multiplication
    Output: DM @ Trans(H) @ H @ DM @ Z
    """

    w = np.dot(DM, Z)  # w  = DM @ Z
    ww = atamult(H, w, 0)  # ww = Trans(H) @ H @ DM @ Z
    w = np.dot(DM, ww)  # w  = DM @ Trans(H) @ H @ DM @ Z
    return w



import logging
from typing import Optional, Any, Callable, Tuple

import numpy as np
from numpy.linalg import norm

from lsqAxSI.default_values import reg_func, jac_calc, X0, MIN_VAL, MAX_VAL
from lsqAxSI.lsq_components import LSQResult, LSQOptions, LSQVars, LSQValues
from lsqAxSI.lsq_trust_region import Trust_region

logger = logging.getLogger(__name__)

EPS = np.finfo('float').eps

"""
Methods:
-------
__init__(reg_func, xstart, user_options, jacob, *varargin):
    Initializes the LSQ optimizer with a residual function, initial guess, options, and Jacobian function.
set_boundaries(lower_bound, upper_bound):
    Sets the lower and upper bounds for the optimization variables.
assert_x_in_bounds():
    Ensures that the current solution is within the specified bounds.
main_loop(jac, *varargin):
    Executes the main optimization loop until convergence or iteration limits are reached.
perform_iteration(jac, *varargin):
    Performs a single iteration of the optimization process, including trust-region updates
    and state advancement.
test_trial_point(new_val):
    Tests whether the proposed trial point satisfies the trust-region acceptance criteria.
determine_trust_region_correction():
    Solves the trust-region subproblem to compute the next step direction and size.
update_delta(new_val):
    Dynamically adjusts the trust-region radius based on the improvement ratio.
advance_to_next_iteration(new_values):
    Updates the optimization state using the evaluated values at the new trial point.
prepare_output():
    Prepares the final output object with optimized parameters, Jacobian, and optimization metadata.
"""


class LSQ:
    """
    Implements a nonlinear least-squares optimization algorithm using a trust-region method.

    The class handles iterative optimization of a set of parameters to minimize the sum of squared residuals,
    incorporating bound constraints, trust-region adjustments, and convergence criteria.

    Attributes:
        func (Callable): The residual function to be minimized. It computes the difference between observed and modeled
                         data given a parameter vector. output dim: (n_frames,)
        options (LSQOptions): Contains user-defined and default options for the optimization process,
                              such as tolerances, maximum iterations, and solver parameters.
        values (LSQValues): Stores the current state of the optimization, including parameter estimates,
                            residuals, Jacobian matrix, gradient, and objective function value.
        vars (LSQVars): Tracks the optimization's iterative variables, such as trust-region radius,
                        proximity to bounds, and iteration counts.
        result (LSQResult): Stores the final outcome of the optimization, including the solution, residuals,
                            and termination status.
        tregion (Trust_region): Handles trust-region calculations for adjusting the step size and direction
                                during the optimization process.
        """

    def __init__(self, reg_func: Callable, xstart: np.ndarray, user_options: dict, jacob: Callable,
                 *varargin: Any) -> None:
        """
        Initialize LSQ for least-squares optimization.

        :param reg_func: Function for the residuals, takes model parameters and data, returns errors. output dim: (n_frames,)
        :param xstart: Initial guess for the model parameters.
        :param user_options: Dictionary of user-specified options for the optimization.
        :param jacob: Jacobian function for the residuals.
        :param varargin: Additional arguments to pass to reg_func and jacob.
        :param tregion (float): Represents the radius of the trust region. This
            dynamically adjusts during the optimization process to balance
            exploration and convergence:
                - If the step improves the objective sufficiently, the trust
                  region may expand.
                - If the step does not improve the objective, the trust
                  region shrinks.
            Ensures the step size stays within bounds for stability and
            convergence efficiency.
        """
        self.func = reg_func
        # TODO: Refael: we can run it one time for all voxels
        self.options = LSQOptions(user_options, xstart)
        self.values = LSQValues(reg_func, xstart, jacob, *varargin)
        self.vars = LSQVars(self.values, self.options)
        self.vars.increment_param("numFunEvals", 3)
        self.result = LSQResult(xstart, self.values.fvec, self.values.A)
        self.tregion = None

    def set_boundaries(self, lower_bound, upper_bound):
        self.vars.set_lower_bound(lower_bound)
        self.vars.set_upper_bound(upper_bound)
        self.vars.assert_bounds()
        self.assert_x_in_bounds()

    def assert_x_in_bounds(self):
        x = self.values.x
        ub = self.vars.ub
        lb = self.vars.lb
        if min(min(ub - x), min(x - lb)) < 0:
            self.values.x = self.vars.compute_startx()

    def main_loop(self, jac, *varargin) -> None:
        """
        Executes the main optimization loop until convergence or iteration limits are reached.

        Parameters
        ----------
        jac: Jacobian function for the residuals.
        varargin: Additional arguments to pass to reg_func and jacob.
        """
        while not self.vars.ex:  # Exit flag status
            self.vars.evaluate_dnewt(self.values.fvec)
            self.update_iteration_variables()
            self.test_for_convergence()
            if not self.vars.ex:
                self.perform_iteration(jac, *varargin)
            self.check_iterations_limit()

    def perform_iteration(self, jac, *varargin) -> None:
        """
        Performs a single iteration of the optimization process, including trust-region updates and state advancement.

        Parameters
        ----------
        jac: Jacobian function for the residuals.
        varargin: Additional arguments to pass to reg_func and jacob.
        """
        reg_func = self.func
        # Computes the next candidate solution x within the trust-region radius
        newx = self.determine_trust_region_correction()
        # Applies a small perturbation to the computed candidate solution to prevent numerical instabilities
        newx = self.perturb(newx)  # TODO: Refael: continue to learn from here
        # This evaluates the residuals, the Jacobian matrix, and the objective function value at the new x
        new_values = LSQValues(reg_func, newx, jac, *varargin)
        # Increments the counter for the number of function evaluations (numFunEvals) by 3 (tracks computational cost).
        self.vars.increment_param("numFunEvals", 3)
        # Determines if the new solution leads to sufficient improvement in the objective function.
        self.test_trial_point(new_values.val)
        # Updates the optimization state using the evaluated values at new x (if accepted)
        self.advance_to_next_iteration(new_values)

    def test_trial_point(self, new_val):
        self.update_ratio(new_val)
        self.update_delta(new_val)

    def update_ratio(self, new_val):
        ss = self.tregion.ss
        dv = self.vars.dv
        g = self.values.grad
        val = self.values.val
        qp = self.tregion.qpval
        aug = 0.5 * np.dot(ss.T, (dv * abs(g)) * ss)
        self.vars.ratio = (0.5 * (new_val - val) + aug) / qp

    def update_delta(self, new_val):
        delta = self.vars.delta
        ratio = self.vars.ratio
        nrmsx = self.vars.nrmsx

        if ratio >= 0.75 and nrmsx >= 0.9 * delta:
            self.vars.delta = min(self.vars.delbnd, 2 * delta)
        elif ratio <= 0.25:
            self.vars.delta = min(nrmsx / 4, delta / 4)
        if np.isinf(new_val):
            self.vars.delta = min(nrmsx / 20, delta / 20)

    def advance_to_next_iteration(self, new_values):
        if new_values.val < self.values.val:
            self.values = new_values
        self.vars.it = self.vars.it + 1
        self.vars.vval[self.vars.it] = self.values.val

    def perturb(self, x, delta=100 * EPS):
        u = self.vars.ub
        l = self.vars.lb
        if (min(abs(u - x)) < delta) or (min(abs(x - l)) < delta):
            upperi = (u - x) < delta
            loweri = (x - l) < delta
            x[upperi] = x[upperi] - delta
            x[loweri] = x[loweri] + delta
        return x

    def update_iteration_variables(self):
        """
        Determines the proximity of the current solution x to its lower and upper bounds.
        Adjusts the gradient based on proximity to bounds (focusing optimization effort on those that are more constrained).
        Tracks the maximum adjusted gradient value for convergence monitoring.
        """
        self.vars.define_dv_and_v(self.values.x, self.values.grad)
        gopt = self.vars.v * self.values.grad
        self.vars.voptnrm[self.vars.it] = norm(gopt, np.inf)

    def test_for_convergence(self):
        """
        Check whether the optimization process has converged based on step size,
        trust-region radius, improvement ratio, and change in the objective function.

        Convergence Criteria:
        1. Trust-region and improvement-based:
           - Step size is small relative to the trust-region radius.
           - Improvement ratio is sufficient (> 0.25).
           - Change in the objective function is below a specified threshold.
        2. Step size-based:
           - The step size is below a strict tolerance.

        Updates:
        - Sets `self.vars.ex` to 1 or 2 to indicate the reason for convergence.
        - Calls `self.result.successful_termination()` if convergence is achieved.
        """

        nrmsx = self.vars.nrmsx
        delta = self.vars.delta
        ratio = self.vars.ratio
        tol1 = self.options.tol1
        val = self.values.val
        it = self.vars.it
        if it > 0:
            # vval[it-2] - Objective function value from the previous iteration, used to compute the change in the objective function
            diff = abs(self.vars.vval[it - 1] - val)
            if (nrmsx < 0.9 * delta) & (ratio > 0.25) & (diff < tol1 * (1 + abs(val))):
                self.result.successful_termination()
                self.vars.ex = 1
            elif (nrmsx < self.options.tol2):
                self.result.successful_termination()
                self.vars.ex = 2

    def check_iterations_limit(self):
        if self.vars.it > self.options.itb:
            self.max_iterations_exceeded()
        if self.vars.numFunEvals > self.options.maxfunevals:
            self.max_iterations_exceeded()

    def max_iterations_exceeded(self):
        self.vars.ex = 4
        self.vars.it = self.vars.it - 1
        self.result.iterations_exceeded()

    def determine_trust_region_correction(self) -> np.ndarray:
        """ Solves the trust-region subproblem to compute the next step direction and size.

        theta controls the size of the trust region: If the optimization is closer to convergence (smaller scaled
        gradient norm - optnorm) theta approaches 0.95, limiting the step size. Otherwise, theta decreases, allowing
        larger exploratory steps.
        """
        iter = self.vars.it
        optnorm = self.vars.voptnrm[iter]
        theta = max(0.95, 1 - optnorm)
        self.tregion = self.trust_region_trial_step(theta)  # computes the optimal step s within the trust region.
        # Updates the current parameter vector x using the computed step s and returns the new candidate solution.
        newx = self.update_tregion_values()  # TODO: Refael: continue to learn from here
        return newx

    def trust_region_trial_step(self, theta):
        x = self.values.x
        delta = self.vars.delta
        options = self.options

        tregion = Trust_region(self.values, self.vars)
        tregion.trial_step(x, theta, options.kmax, options.pcflags, options.pcgtol, delta)
        return tregion

    def update_tregion_values(self):
        iter = self.vars.it
        posdef = self.tregion.posdef
        if posdef == 0:
            posdef = self.vars.vpos[iter]
        self.vars.vpos[iter + 1] = posdef
        self.vars.nrmsx = norm(self.tregion.ss)
        self.vars.vpcg[iter + 1] = self.tregion.pcgit
        newx = self.values.x + self.tregion.s
        return newx

    def prepare_output(self):
        self.result.jacob = self.values.A
        self.set_output_values()
        self.result.x = self.values.x
        self.result.fvec = self.values.fvec

    def set_output_values(self):
        it = self.vars.it
        output = self.result.output
        output['firstorderopt'] = self.vars.voptnrm[it]
        output['iterations'] = it
        output['funcCount'] = self.vars.numFunEvals
        output['cgiterations'] = np.sum(self.vars.vpcg)


def snls(funfcn: Callable, xstart: np.ndarray, l: np.ndarray, u: np.ndarray, options: dict,
         jac: Callable, *varargin: Any) -> LSQResult:
    """
    Solves a nonlinear least squares problem with bound constraints.

    :param funfcn: The objective function to minimize.
    :param xstart: Initial guess for the solution. Shape: (n_params,), in our case (2,).
    :param l: Lower bounds for the parameters. Shape: (n_params,), in our case (2,).
    :param u: Upper bounds for the parameters. Shape: (n_params,), in our case (2,).
    :param options: Optimization options such as tolerances and maximum iterations.
    :param jac: Function to compute the Jacobian matrix of the objective function.
    :param varargin: Additional arguments passed to `funfcn` and `jac`.
    :return: LSQResult object with the result of the optimization process.
    """
    lsq_component = LSQ(funfcn, xstart, options, jac, *varargin)
    lsq_component.set_boundaries(l, u)
    lsq_component.main_loop(jac, *varargin)
    lsq_component.prepare_output()

    return lsq_component.result


def nonlinear_least_squares(reg_func: Callable = reg_func, x0: np.ndarray = X0,
                            bounds: Tuple[np.ndarray, np.ndarray] = (MIN_VAL, MAX_VAL), jac: Callable = jac_calc,
                            ftol: float = 1e-6, xtol: float = 1e-6, diff_step: float = 1e-3,
                            max_nfev: int = 20000, args: Optional[Tuple[Any, ...]] = None) -> LSQResult:
    """
    Perform nonlinear least-squares optimization using a customized solver.

    This function minimizes the sum of squares of residuals defined by the
    provided regression function `reg_func`.

    Parameters:
    ----------
    reg_func : callable
        The regression function that computes the residuals (for each frame). It must have the signature:
        `reg_func(x, *args)` where `x` is the parameter vector to optimize, and
        `*args` are additional arguments.

    x0 : array-like
        Initial guess for the parameters to be optimized. The dimension of
        `x0` corresponds to the number of parameters in the model (2 paramters in the default case).

    bounds : tuple of array-like, optional
        Lower and upper bounds for each parameter in `x0`.
        Format: `(lower_bounds, upper_bounds)`, where both are arrays of the
        same dimension as `x0`. Default is `None`, meaning no bounds are imposed.

    jac : callable, optional
        A function to compute the Jacobian matrix of the residuals. If `None`,
        the Jacobian is approximated using finite differences. The Jacobian function
        must have the signature `jac(x, *args)` and return a 2D array of size
        `(n_frames, 2)`.

    ftol : float, optional
        Tolerance for the cost function value. Iterations will stop when the
        relative change in the cost function is less than `ftol`. Default is `1e-6`.

    xtol : float, optional
        Tolerance for parameter updates. Iterations will stop when the relative
        change in the parameter vector is less than `xtol`. Default is `1e-6`.

    diff_step : float, optional
        Step size for finite-difference approximation of the Jacobian if
        `jac` is not provided. Default is `1e-3`.

    max_nfev : int, optional
        Maximum number of function evaluations before terminating the optimization.
        Default is `20,000`.

    args : tuple, optional
        Additional arguments to pass to the `reg_func` and `jac`. These are passed
        as `*args`.

    Returns:
    -------
    result : LSQResult object
        An object containing the optimized parameter vector, cost function value,
        Jacobian matrix, and additional optimization details.

    Notes:
    ------
    This function is a wrapper for a custom solver, which uses trust-region methods
    for robust nonlinear least-squares optimization. It extends standard least-squares
    techniques with additional flexibility and control over parameters.
    """

    # xstart = np.array(x0, dtype='float')   # Refael: it is already np.array
    # l = np.array(bounds[0], dtype='float') # Refael: it is already np.array
    # u = np.array(bounds[1], dtype='float') # Refael: it is already np.array

    options = {'TolFun': ftol, 'TolX': xtol, 'MaxFunEvals': max_nfev, 'diff_step': diff_step}
    varargin = []
    for i in range(len(args) - 1):
        varargin.append(np.array(args[i]).flatten())
    varargin.append(args[-1])

    # result = snls(reg_func, xstart, l, u, options, jac, *varargin) # Refael
    result = snls(reg_func, x0, bounds[0], bounds[1], options, jac, *varargin)  # Refael
    return result



import logging
from typing import Any, Callable, Tuple

import numpy as np
from numpy.linalg import norm

from lsqAxSI.matrix_operations import atamult

logger = logging.getLogger(__name__)

EPS = np.finfo('float').eps
EMPTY_ARRAY = np.array([])
INF = np.inf
defaultopt = {
    'ActiveConstrTol': np.sqrt(EPS),
    'PrecondBandWidth': INF,
    'tolPCG': 0.1,
    'MaxPCGIter': 'max(1,floor(numberofvariables/2))',
    "TypicalX": 'ones(numberofvariables,1)',
    "maxfunevals": '100*numberofvariables',
    'MaxIter': 20000
}


class LSQResult:
    """
    Represents the result of a least-squares optimization process.

    Attributes:
        x (np.ndarray): The final optimized parameter vector.
        fvec (np.ndarray): The residual vector at the solution.
        jacob (np.ndarray): The Jacobian matrix at the solution.
        exitflag (int): Status flag indicating success or failure (1 for success, 0 for failure).
        output (Dict[str, str]): Dictionary containing details about the optimization algorithm.
        msg (str): A message describing the result or status of the optimization.
    """

    def __init__(self, xstart: np.ndarray, fvec: np.ndarray, jacob: np.ndarray) -> None:
        """
        Initializes an LSQResult instance.

        Args:
            xstart (np.ndarray): Initial parameter vector.
            fvec (np.ndarray): Residual vector.
            jacob (np.ndarray): Jacobian matrix.
        """
        self.x = xstart
        self.fvec = fvec
        self.jacob = jacob
        self.exitflag = 0
        self.output = {'algorithm': 'trust-region-reflective'}
        self.msg = ""

    def successful_termination(self):
        self.exitflag = 1
        self.msg = "Optimization terminated successfully"

    def iterations_exceeded(self):
        self.exitflag = 0
        self.msg = "Maximum number of iterations exceeded"


class LSQOptions():
    """
    Class for managing and validating the options for the least squares optimization.
    This includes user-specified options and default options for the solver.

    Attributes:
        user_options (dict): Dictionary of user-provided options.
        default_options (dict): Default options for the solver.
        active_tol (float): Tolerance for active constraints.
        gradflag (bool): Whether the Jacobian is active.
        typx (np.ndarray): Typical values for the optimization variables.
        pcflags (float): Preconditioning bandwidth flag.
        tol2 (float): Tolerance for x values.
        tol1 (float): Tolerance for function values.
        itb (int): Maximum number of iterations for the solver.
        maxfunevals (int): Maximum number of function evaluations.
        pcgtol (float): Tolerance for preconditioned conjugate gradient.
        kmax (int): Maximum iterations for preconditioned conjugate gradient.
        numberOfVariables (int): Number of variables to optimize.
    """

    def __init__(self, user_options: dict, xstart: np.ndarray):
        """
        Initialize LSQOptions with user-specified options and default settings.

        Args:
            user_options (dict): A dictionary of user-defined options.
            xstart (np.ndarray): Initial guess for the optimization variables.
        """

        self.user_options = user_options
        self.default_options = defaultopt
        # Refael. we can do it simpler (do it for all keys) - can delete self.optimget method:
        # TODO: test it
        self.default_options.update(user_options)
        self.active_tol = self.default_options.get('ActiveConstrTol')  # get value or None if not exists
        self.gradflag = self.default_options.get('Jacobian') == 'on'
        self.typx = self.default_options.get('TypicalX')
        self.pcflags = self.default_options.get('PrecondBandWidth')
        self.tol2 = self.default_options.get('TolX')
        self.tol1 = self.default_options.get('TolFun')
        self.itb = self.default_options.get('MaxIter')
        self.maxfunevals = self.default_options.get('MaxFunEvals')
        self.pcgtol = self.default_options.get('tolPCG')
        self.kmax = self.default_options.get('MaxPCGIter')
        # self.active_tol = self.optimget('ActiveConstrTol')
        # self.gradflag = self.optimget('Jacobian') == 'on'
        # self.typx = self.optimget('TypicalX')
        # self.pcflags = self.optimget('PrecondBandWidth')
        # self.tol2 = self.optimget('TolX')
        # self.tol1 = self.optimget('TolFun')
        # self.itb = self.optimget('MaxIter')
        # self.maxfunevals = self.optimget('MaxFunEvals')
        # self.pcgtol = self.optimget('tolPCG')
        # self.kmax = self.optimget('MaxPCGIter')
        self.numberOfVariables = xstart.shape[0]
        self.assert_options()

    # Refael - delete it
    # def optimget(self, key:str, default_value=None) -> Any:
    #     if key in self.user_options.keys():
    #         return self.user_options[key]
    #     if key in self.default_options.keys():
    #         return self.default_options[key]
    #     return default_value

    def assert_options(self):
        """Validate that the options are correct and raise exception if necessary."""
        if self.numberOfVariables == 0:
            # TODO: Refael: need to through exception
            print("Warning: Number of variables must be positive")
        if self.pcgtol <= 0:
            self.pcgtol = 0.1
        self.assert_attribute("typx", 'ones(numberofvariables,1)', np.ones(self.numberOfVariables))
        self.assert_attribute("kmax", 'max(1,floor(numberofvariables/2))', max(1, self.numberOfVariables // 2))
        self.assert_attribute("maxfunevals", '100*numberofvariables', 100 * self.numberOfVariables)

    def assert_attribute(self, attribute: str, pattern: str, default_value: any):
        """
        Ensure that a given attribute matches the expected pattern,
        and set it to a default value if necessary.

         Args:
             attribute (str): The attribute to check.
             pattern (str): The expected pattern for the attribute.
             default_value (any): The default value to set if the pattern matches.
         """
        current_value = getattr(self, attribute)
        if isinstance(current_value, str):
            if current_value.lower() == pattern:
                setattr(self, attribute, default_value)
            else:
                # TODO: Refael: need to through exception
                print(f"Option {attribute} must be integer value if not the default")


class LSQValues():
    """
    Class representing the computed values for the least squares optimization problem.

    Attributes:
        x: Current estimate of the optimization variables. dim: (2, )
        fvec: Residual vector computed by the regularization function. dim: (n, 1).
        A: Jacobian matrix of the residuals with respect to the optimization variables. dim: (n, 2).
        grad: Gradient of the objective function. dim: (2, 1).
        val: Value of the objective function (sum of squared residuals). (float).
    """

    def __init__(self, reg_func: Callable, x: np.ndarray, jac: Callable, *varargin: Tuple[Any]) -> None:
        """
        Initialize LSQValues with computed residuals, Jacobian, and other derived quantities.

        Parameters:
            reg_func: Function to compute the residuals (fvec) given x and additional arguments.
            x: Current estimate of the optimization variables. dim: (2, )
            jac: Function to compute the Jacobian matrix of the residuals.
            varargin: Additional arguments to be passed to reg_func and jac.

        Attributes:
            x: Current estimate of the optimization variables. dim: (2, )
            fvec: Residual vector computed by the regularization function. dim: (n_frames, 1).
            A: Jacobian matrix of the residuals with respect to the optimization variables. dim: (n_frames, 2).
            grad: Gradient of the objective function. dim: (2, 1).
            val: Value of the objective function (sum of squared residuals). (float).
        """
        self.x = x
        self.fvec = reg_func(x, *varargin)
        self.A = jac(x, *varargin)
        self.grad = atamult(self.A, self.fvec, -1)
        self.val = np.dot(self.fvec, self.fvec)
        self.assert_fvec()

    def assert_fvec(self):
        # Refael: the dimension of x is (2, ) and the dimension of fvec is (n_frame, ).
        if self.fvec.shape[0] < self.x.shape[0]:
            # TODO: Refael: need to through exception ???
            print("the number of equations must not be less than n")


class LSQVars:
    """
    Represents the variables and parameters used during the iterative
    optimization process in a least-squares problem.

    Attributes:
        n (int): Number of optimization variables (dimensions of `x`).
        lb (np.ndarray): Lower bounds for the optimization variables.
        ub (np.ndarray): Upper bounds for the optimization variables.
        dnewt (np.ndarray): Newton direction for updates.
        it (int): Current iteration number.
        numFunEvals (int): Number of function evaluations.
        numGradEvals (int): Number of gradient evaluations.
        ex (int): Exit flag status.
        vpcg (np.ndarray): Vector storing trust-region PCG iterations.
        vpos (np.ndarray): Vector storing positivity flags for trust-region.
        vval (np.ndarray): Vector storing objective function values per iteration.
        voptnrm (np.ndarray): The infinity norm of the scaled gradient (by self.v) at the current iteration.
        delta (float): Trust-region radius, which limits the size of parameter updates in trust-region optimization methods.
        nrmsx (float): Norm of step size. presenting how much the parameters x have changed in the current iteration.
        ratio (float): Ratio of the actual improvement in the objective function to the predicted improvement, used to
                       evaluate whether the step taken was effective.
        v (np.ndarray): Vector representing proximity of `x` to the bounds (`lb` and `ub`). dim: (2,).
        dv (np.ndarray): A binary mask indicating whether each parameter is actively constrained
                         by the bounds. Values are `1` if constrained and `0` otherwise.. dim: (2,).
        delbnd (float): Upper bound for trust-region delta.
    """

    def __init__(self, values: LSQValues, options: LSQOptions):
        self.n = values.x.shape[0]  # param
        self.lb = EMPTY_ARRAY  # param
        self.ub = EMPTY_ARRAY  # param
        self.dnewt = EMPTY_ARRAY
        self.it = 0  # counters
        self.numFunEvals = 0  # counters
        self.numGradEvals = 0  # counters
        self.ex = 0
        self.vpcg = np.zeros(options.itb)  # vector of trustregion.pcgit
        self.vpos = np.ones(options.itb)  # vector of posdef
        self.vval = np.zeros(options.itb)  # vector of vals
        self.vval[self.it] = values.val
        self.voptnrm = np.zeros(options.itb)  # vector
        self.delta = 10
        self.nrmsx = 1
        self.ratio = 0
        self.v = EMPTY_ARRAY
        self.dv = EMPTY_ARRAY
        self.delbnd = max(100 * norm(values.x), 1)  # param

    def increment_param(self, name, add):
        setattr(self, name, getattr(self, name) + add)

    def set_lower_bound(self, lower_bound):
        if lower_bound.size == 0:
            lower_bound = -INF * np.ones(self.n)
        lower_bound[lower_bound <= -1e10] = -INF
        self.lb = lower_bound

    def set_upper_bound(self, upper_bound):
        if upper_bound.size == 0:
            upper_bound = INF * np.ones(self.n)
        upper_bound[upper_bound >= 1e10] = INF
        self.ub = upper_bound

    def assert_bounds(self):
        if np.any(self.ub == self.lb):
            print("equal upper and lower bound not permitted")
            exit()
        elif min(self.ub - self.lb) <= 0:
            print("inconsistent bounds")
            exit()

    def compute_startx(self):
        '''startx returns centered point'''
        arg1 = (self.ub < INF) & (self.lb == -INF)
        arg2 = (self.ub == INF) & (self.lb > -INF)
        arg3 = (self.ub < INF) & (self.lb > -INF)
        arg4 = (self.ub == INF) & (self.lb == -INF)
        return self.set_startx_values(arg1, arg2, arg3, arg4)

    def set_startx_values(self, arg1, arg2, arg3, arg4):
        xstart = np.zeros(self.n)
        w = np.maximum(abs(self.ub), np.ones(self.n))
        ww = np.maximum(abs(self.lb), np.ones(self.n))

        xstart[arg1] = self.ub[arg1] - 0.5 * w[arg1]
        xstart[arg2] = self.lb[arg2] + 0.5 * ww[arg2]
        xstart[arg3] = (self.ub[arg3] + self.lb[arg3]) / 2
        xstart[arg4] = 1
        return xstart

    def define_dv_and_v(self, x: np.ndarray, grad: np.ndarray) -> None:
        """
        Defines the proximity (`v`) and directional adjustment indicator (`dv`)
        for the optimization parameters based on the current gradient and bounds.

        Args:
            x (np.ndarray): Current parameter values.
            grad (np.ndarray): Gradient vector of the objective function.

        Attributes Updated:
            -   `self.v` measures how close the current parameters (x) are to the bounds (lb and ub)
            -   `self.dv` is a binary vector indicating whether adjustments are needed for the bounds.
                    dv[i] = 1 if the corresponding parameter (x[i]) is actively interacting with a bound.
                    dv[i] = 0 if the parameter is unconstrained.
        """
        arg1 = (grad < 0) & (self.ub < INF)
        arg2 = (grad >= 0) & (self.lb > -INF)
        arg3 = (grad < 0) & (self.ub == INF)
        arg4 = (grad >= 0) & (self.lb == -INF)

        self.define_v(x, arg1, arg2, arg3, arg4)
        self.define_dv(arg1, arg2, arg3, arg4)

    def define_v(self, x: np.ndarray, arg1: np.ndarray, arg2: np.ndarray, arg3: np.ndarray, arg4: np.ndarray) -> None:
        v = np.zeros(self.n)
        v[arg1] = x[arg1] - self.ub[arg1]
        v[arg2] = x[arg2] - self.lb[arg2]
        v[arg3] = -1
        v[arg4] = 1
        self.v = v

    def define_dv(self, arg1: np.ndarray, arg2: np.ndarray, arg3: np.ndarray, arg4: np.ndarray) -> None:
        dv = np.zeros(self.n)
        dv[arg1] = 1
        dv[arg2] = 1
        dv[arg3] = 0  # maybe redundant
        dv[arg4] = 0  # maybe redundant
        self.dv = dv

    def evaluate_dnewt(self, fvec):
        if np.any(np.isinf(fvec)):
            # TODO: Refael: what we need to do with this message ???
            print("user function is returning inf or NaN values")
        # TODO: Refael:
        # 1) When is `fvec` 2D? Does it occur at any point in the algorithm?
        # 2) What is `self.vars`? Should it be `self.dnewt` instead?
        # 3) If `fvec` is 1D, when (if at all) we update `dnewt`?
        if fvec.ndim == 2 and fvec.shape[1] == 2:
            self.vars.dnewt = fvec[:, 1]



import numpy as np

#########################################################################################################
# Default reg_func and jac_func functions
#########################################################################################################

X0 = np.asarray([0.5, 5000.0])
MIN_VAL = np.zeros(2)
MAX_VAL = np.array([1.0, 20000.0])


# Placeholder regression function
def reg_func(x: np.ndarray, ydata: np.ndarray, pixpredictH: np.ndarray, pixpredictR: np.ndarray,
             pixpredictCSF: np.ndarray, prcsf: float) -> np.ndarray:
    """
    Regression function for nonlinear least squares.

    Computes the error between observed data (`ydata`) and predicted data based on
    a model combining hindered, restricted, and CSF signal components, using a set
    of parameters.

    :param x: Array of parameters for the model, where:
              - x[0]: Fraction of the hindered component.
              - x[1]: Scaling factor for the predicted signal.
    :param ydata: Observed signal data. Shape: (n_frames,).
    :param pixpredictH: Predicted hindered component of the signal. Shape: (n_frames,).
    :param pixpredictR: Predicted restricted component of the signal. Shape: (n_frames,).
    :param pixpredictCSF: Predicted CSF (cerebrospinal fluid) component of the signal. Shape: (n_frames,).
    :param prcsf: Fraction of the CSF component in the signal.
    :return: The residual error between the observed (`ydata`) and the predicted signal.
             Shape: (n_frames,).
    """
    # xt = 1 - x[0] - prcsf
    # newdata = x[1] * (x[0] * pixpredictH + xt * pixpredictR + prcsf * pixpredictCSF)
    newdata = x[1] * (x[0] * pixpredictH + (1 - x[0] - prcsf) * pixpredictR + prcsf * pixpredictCSF)
    err = newdata - ydata
    return err


# Placeholder jacobian function
def jac_calc(x, ydata, pixpredictH, pixpredictR, pixpredictCSF, prcsf) -> np.ndarray:
    """ jacobian matrix calculation for nonlinear least squares (reg_func function). dim: (n_frames, 2) """
    jac = np.zeros([len(ydata), 2])
    jac[:, 0] = x[1] * (pixpredictH - pixpredictR)
    jac[:, 1] = x[0] * pixpredictH + (1 - x[0] - prcsf) * pixpredictR + prcsf * pixpredictCSF
    return jac

#########################################################################################################



import os

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

requires = ['numpy', 'scipy', 'scikit-learn']

setuptools.setup(
    # required
    name="lsqAxSI",
    version="0.0.6",
    # optional
    author="Hila Gast",
    author_email="",
    description="Nonlinear least-squares optimization for AxSI",
    long_description=long_description,  # This is shown on the package detail package on the Pypi
    long_description_content_type="text/markdown",  # text/plain, text/x-rst (for reStructuredText), text/markdown
    url="",  # github

    classifiers=[  # Meta-data. Examples: https://pypi.org/classifiers/
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=requires,  # Install dependencies
    scripts=[
        'scripts/lsq_AxSI.py',
    ],

    packages=setuptools.find_packages(),

    data_files=[('docs', ['docs/usage.rst']), ],

    package_data={
        # If any package contains *.txt or *.rst files, include them:
        "": ["*.rst"],
    },
    #
    # # For sdist only:
    # # True: include only files from MANIFEST.in file (not from package_data).
    # # False: include files from MANIFEST.in file AND from package_data
    # include_package_data=False,
    #
    # tests_require=requires + ['nose2'], # Install requirements when you run: python setup.py test
    # test_suite='nose2.collector.collector',
)
